#include <stdio.h>
#include <stdlib.h>
#include "tags.h"

#define PART_ENTRIES 256

#define S_COL	1
#define S_SIZE	2
#define S_U1	4
#define S_U2	8
#define S_U3	16
#define S_TT	32
#define S_EM	64
#define S_S	128
#define S_I	256
#define S_B	512
#define S_EMONLY 1024

typedef struct PartInfo {
  BlockNode *start;
  BlockNode *end;
  int pos;
  int cost;  //contains length info
  short minstate;
  short placeholder;
} PartInfo;

/* This section of the code is used to partition
 * out segments of the input code so that
 * overlapping regions can be handled */

int CostFunc(const void *a, const void *b)
{
  if (((PartInfo*)a)->start == NULL)
    return 1;   // Shift unused entries to the bottom
  return (((PartInfo*)a)->cost-((PartInfo*)b)->cost);
}

short GetPState(short s)
{
  short p=S_EMONLY;
  if (s & M_B) p|=S_B;
  if (s & M_I) p|=S_I;
  if (s & M_S) p|=S_S | S_EM; // yes, set both here
  if (s & M_EM) p=(p|S_EM)&(~S_EMONLY);
  if (s & M_TT) p|=S_TT;
  if ((s&M_U) == 0x0300) p|=S_U3|S_U2|S_U1;
  if ((s&M_U) == 0x0200) p|=S_U2|S_U1;
  if ((s&M_U) == 0x0100) p|=S_U1;
  if (s & M_SIZE) p|=S_SIZE;
  if (s & M_COL) p|=S_COL;
  return p;
}

//Returns the cost of making a partition
int GetCost(BlockNode *s, BlockNode *e, short minstate)
{
  int cost=0;
  int x;
  short instate, outstate;
  //Check to make sure these are valid blocks...
  if (s == NULL)
    return 100;
  //These are the states that are true through the boundaries
  instate=GetPState(s->state)&GetPState(s->change^s->state);
  outstate=GetPState(e->state)&GetPState(e->change^e->state);
  //First check the upper 8 attributes
  if (minstate & S_EMONLY) minstate|=S_EM;
  for (x=4;x&1023;x<<=1)
    {
      if (!(x & minstate))	// if it is not always set
	{
	  if (x & instate) cost++;
	  if (x & outstate) cost++;
	}
    }
  // Okay, now for size and color
  // Size:
  if (!(minstate & S_SIZE)) //if size was not always set
    {
      if (instate & S_SIZE) cost++;
      if (outstate & S_SIZE) cost++;
    }
  // Color:
  if (!(minstate & S_COL)) //if size was not always set
    {
      if (instate & S_COL) cost++;
      if (outstate & S_COL) cost++;
    }
/*   else */
/*     { */
/*       if ((s->state & M_COL)!=(e->state & M_COL)) cost++; */
/*     } */
  return cost;
}

//Load a partition into the PartInfo structure
void PartLoad(PartInfo *pi, int *pos, PartInfo *npi)
{
  if (*pos==PART_ENTRIES)
    {
      qsort(pi, PART_ENTRIES, sizeof(PartInfo), CostFunc);
      *pos=PART_ENTRIES/2;
    }
  memcpy(&(pi[*pos]), npi, sizeof(PartInfo));
  *pos+=1;
}

//Find places to partition the program
int PartFind(Block *bl, PartInfo *pi)
{
  BlockNode *c, *start[10];
  PartInfo lpi;
  int pi_pos=0, x, p, pos=0, startpos[10], t1;
  short new_pstate, minstate[10];
  for(x=0;x<10;x++)
    start[x]=NULL;
  // Begin
  if (bl->start == bl->end)
    return 0;	// termination case
  c=bl->start;
  while (1)
    {
      // Find out what changed
      new_pstate=(~GetPState(c->change^c->state))&(GetPState(c->state));
      p=2;
      for (x=4;x&1023;x<<=1)
	{
	  if (new_pstate & x) // if event went high
	    {
	      minstate[p]=GetPState(c->state);
	      start[p]=c;
	      startpos[p]=pos;
	    }
	  p+=1;
	}
      new_pstate=(GetPState(c->change^c->state))&(~GetPState(c->state));
      p=2;
      for (x=4;x&1023;x<<=1)
	{
	  if (new_pstate & x) // if event went low
	    {
	      t1=GetCost(start[p], c, minstate[p]);
	      lpi.start=start[p];
	      lpi.end=c;
	      lpi.pos=startpos[p];
	      lpi.cost=(t1<<24)|(0x00FFFFFF &(startpos[p]-pos));
	      lpi.minstate=minstate[p];
	      if ((lpi.start!=bl->start)||(lpi.end!=bl->end))
/* 		{ */
/* 		  if ((t1)||(pos-startpos[p]>1)) */
/* 		    PartLoad(pi, &pi_pos, &lpi);  */
/* 		} */
		{ if (t1) PartLoad(pi, &pi_pos, &lpi); }
/* 	      else */
/* 		fprintf(stderr, "Caught full length partition!\n"); */
	    }
	  p+=1;
	}
      // Color
      if (c->change & M_COL)
	{
	  if ((c->change^c->state) & M_COL)
	    {
	      t1=GetCost(start[0], c, minstate[0]);
	      lpi.start=start[0];
	      lpi.end=c;
	      lpi.pos=startpos[0];
	      lpi.cost=(t1<<24)|(0x00FFFFFF &(startpos[0]-pos));
	      lpi.minstate=minstate[0];
	      if ((lpi.start!=bl->start)||(lpi.end!=bl->end))
		{
		  if ((t1)||(pos-startpos[0]>1))
		    PartLoad(pi, &pi_pos, &lpi); 
		}
/* 	      else */
/* 		fprintf(stderr, "Caught full length partition!\n"); */
/* 	      if (t1) */
/* 		PartLoad(pi, &pi_pos, &lpi); */
	    }
	  if (c->state & M_COL)
	    {
	      minstate[0]=GetPState(c->state);
	      start[0]=c;
	      startpos[0]=pos;
	    }
	}
      // Size
      if (c->change & M_SIZE)
	{
	  if ((c->change^c->state) & M_SIZE)
	    {
	      t1=GetCost(start[1], c, minstate[1]);
	      lpi.start=start[1];
	      lpi.end=c;
	      lpi.pos=startpos[1];
	      lpi.cost=(t1<<24)|(0x00FFFFFF &(startpos[1]-pos));
	      lpi.minstate=minstate[1];
	      if ((lpi.start!=bl->start)||(lpi.end!=bl->end))
		{
		  if ((t1)||(pos-startpos[1]>1))
		    PartLoad(pi, &pi_pos, &lpi);
		}
/* 	      else */
/* 		fprintf(stderr, "Caught full length partition!\n"); */
/* 	      if (t1) */
/* 		PartLoad(pi, &pi_pos, &lpi); */
	    }
	  if (c->state & M_SIZE)
	    {
	      minstate[1]=GetPState(c->state);
	      start[1]=c;
	      startpos[1]=pos;
	    }
	}
      // Update the minstate bits
      new_pstate=GetPState(c->state);
      for (x=0;x<10;x++)
	minstate[x]&=new_pstate;
      if (c==bl->end)
	return 0;
      c=c->next;
      pos+=1;
    }
}

// Create and execute a partition
int PartRun(PartInfo *pi, Block *bl2)
{
  short status=0;
  Block bl;
/*   fprintf(stderr, "Entry: pos %d, length %d, cost %d, boundary %04x\n", */
/* 	  pi->pos, -(pi->cost | 0xFF000000), pi->cost>>24, pi->minstate); */
  // Step one: Figure out what the partition status should be
  if (pi->minstate & S_B) status|=M_B;
  if (pi->minstate & S_I) status|=M_I;
  if (pi->minstate & S_TT) status|=M_TT;
  if (pi->minstate & S_U3) status|=0x0300;
  else if (pi->minstate & S_U2) status|=0x0200;
  else if (pi->minstate & S_U1) status|=0x0100;
  if (pi->minstate & S_S) status|=M_S;
  else if (pi->minstate & S_EM) status|=M_EM;
  // Size
  if (pi->minstate & S_SIZE)
    {
      //If we agree internally, use that
      if ((pi->start->state &M_SIZE)==
	  ((pi->end->change^pi->end->state)&M_SIZE))
	status|=(pi->start->state&M_SIZE);
      //If one side lacks a transition, use that
      else if ((pi->start->change&M_SIZE)==0)
	status|=(pi->start->state&M_SIZE);
      else if ((pi->end->change&M_SIZE)==0)
	status|=(pi->end->state&M_SIZE);
      //Crap, there's no obvious advantages here
      //Pick one and hope this doesn't come up often...
      else status|=pi->start->state & M_SIZE;
    }
  // Color
  if (pi->minstate & S_COL)
    {
      //If we agree internally, use that
      if ((pi->start->state &M_COL)==
	  ((pi->end->change^pi->end->state)&M_COL))
	status|=(pi->start->state&M_COL);
      //If one side lacks a transition, use that
      else if ((pi->start->change&M_COL)==0)
	status|=(pi->start->state&M_COL);
      else if ((pi->end->change&M_COL)==0)
	status|=(pi->end->state&M_COL);
      //Crap, there's no obvious advantages here
      //Pick one and hope this doesn't come up often...
      else status|=pi->start->state & M_COL;
    }
/*   fprintf(stderr, "Using %04x as status\n", status); */
  // Step two: Add the boundary block nodes
  block_insert(pi->start, status);
  block_insert(pi->end, status);
  if (pi->start==bl2->start)
    {
/*       fprintf(stderr, "Caught part begin error!\n"); */
      bl2->start=pi->start->prev;
    }
  // Step Three: Optimize the partition
  bl.start=pi->start;
  bl.end=pi->end->prev;
  return reduce(&bl);
}

// Top level partitioning function
int Partition(Block* bl)
{
  PartInfo pi[PART_ENTRIES];
  int x, y, pos;
  for (x=0;x<PART_ENTRIES;x++)
    pi[x].start=NULL;
  // Search for good partitioning points
/*   fprintf(stderr,"Begin PartFind\n"); */
  x=PartFind(bl, pi);
/*   fprintf(stderr,"End PartFind\n"); */
  // *** Eliminate unusable partitions (because of overlaps)
  // Sort by cost
  qsort(pi, PART_ENTRIES, sizeof(PartInfo), CostFunc);
  // Scan each entry
  for (x=0;x<PART_ENTRIES;x++)
    {
      if (pi[x].start!=NULL)
	{
	  PartRun(&pi[x], bl);
	  pos=pi[x].pos-(pi[x].cost | 0xFF000000);
	  // Compare this entry to the others
	  for (y=x;y<PART_ENTRIES;y++)
	    if (pi[y].pos <= pos)
	      pi[y].start=NULL;
	}
    }
  return 0;
  // Now do the partitions
  //fill in later
  
}
